/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define FE_SDO_DEBUG	FALSE
#define	FE_SDO_PROFILE	(FE_CODEGEN==FE_PROFILE)

#include "operator/operator.pmh"

namespace fe
{

void SubdivideOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("partitionAttr")="hull_primitive";
	catalog<String>("partitionAttr","label")="Partition Attr";
	catalog<String>("partitionAttr","hint")=
			"Primitive string attribute created on output"
			" distinguishing the original primitives.";

	catalog<bool>("recomputeNormals")=true;
	catalog<String>("recomputeNormals","label")="Recompute Normals";

	catalog<I32>("depth")=1;
	catalog<I32>("depth","high")=8;
	catalog<I32>("depth","max")=32;
	catalog<String>("depth","label")="Depth";
	catalog<String>("depth","hint")=
			"Number of subdivision iterations.";

	catalog<I32>("flatDepth")=0;
	catalog<I32>("flatDepth","high")=8;
	catalog<I32>("flatDepth","max")=32;
	catalog<String>("flatDepth","label")="Flat Depth";
	catalog<String>("flatDepth","hint")=
			"Number of iterations to linearly interpolate (at finest end).";

	catalog< sp<Component> >("Input Surface");

	//* don't copy input to output
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";

	m_spSurfaceNormalOp=registry()->create("OperatorSurfaceI.SurfaceNormalOp");
}

void SubdivideOp::bind(sp<SignalerI> a_spSignalerI,sp<Layout> a_spLayout)
{
	OperatorSurfaceCommon::bind(a_spSignalerI,a_spLayout);

	sp<OperatorSurfaceCommon> spCommon=m_spSurfaceNormalOp;
	if(spCommon.isValid())
	{
		spCommon->bind(a_spSignalerI,a_spLayout);
	}
}

void SubdivideOp::clearData(void)
{
	m_facePoint.clear();
	m_faceSum.clear();
	m_faceCount.clear();
	m_midpointSum.clear();
	m_midpointCount.clear();
	m_boundarySum.clear();
	m_boundaryCount.clear();

	m_edgeMap.clear();
}

void SubdivideOp::handle(Record& a_rSignal)
{
	m_spProfiler=new Profiler("SubdivideOp");
	m_spProfileAccess=new Profiler::Profile(m_spProfiler,"Access");
	m_spProfileCopy=new Profiler::Profile(m_spProfiler,"Copy");
	m_spProfileResize=new Profiler::Profile(m_spProfiler,"Resize");
	m_spProfileSlot=new Profiler::Profile(m_spProfiler,"Slot");
	m_spProfileFace=new Profiler::Profile(m_spProfiler,"Face");
	m_spProfileFacePoint=new Profiler::Profile(m_spProfiler,"FacePoint");
	m_spProfileVertSize=new Profiler::Profile(m_spProfiler,"VertSize");
	m_spProfileEdgeMap=new Profiler::Profile(m_spProfiler,"EdgeMap");
	m_spProfileEdgePoint=new Profiler::Profile(m_spProfiler,"EdgePoint");
	m_spProfileVertex=new Profiler::Profile(m_spProfiler,"Vertex");
	m_spProfileAppend=new Profiler::Profile(m_spProfiler,"Append");
	m_spProfileEdge=new Profiler::Profile(m_spProfiler,"Edge");
	m_spProfilePoint=new Profiler::Profile(m_spProfiler,"Point");
	m_spProfileNormal=new Profiler::Profile(m_spProfiler,"Normal");

	m_spProfiler->begin();
	m_spProfileAccess->start();

	const String partitionAttr=catalog<String>("partitionAttr");
	const BWORD recomputeNormals=catalog<bool>("recomputeNormals");
	const I32 depth=catalog<I32>("depth");
	const I32 flatDepth=catalog<I32>("flatDepth");

	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	if(depth<1)
	{
		if(!access(m_spInputAccessible,"Input Surface")) return;

		m_spOutputAccessible->copy(m_spInputAccessible);
	}

	for(I32 iteration=0;iteration<depth;iteration++)
	{
		m_flat=(iteration>=depth-flatDepth);

		if(iteration)
		{
			//* copy prior output to a new input buffer
			m_spInputAccessible=registry()->create("SurfaceAccessibleI");

			sp<Scope> spScope=registry()->create("Scope");
			m_spInputAccessible->bind(spScope);

			m_spProfileCopy->replace(m_spProfileAccess);

			m_spInputAccessible->copy(m_spOutputAccessible);

			m_spProfileAccess->replace(m_spProfileCopy);

			//* vacate prior output
			m_spOutputAccessible->clear();
		}
		else
		{
			if(!access(m_spInputAccessible,"Input Surface")) return;
		}

		sp<SurfaceAccessorI> spOutputVertices;
		if(!access(spOutputVertices,m_spOutputAccessible,
				e_primitive,e_vertices)) return;

		sp<SurfaceAccessorI> spOutputPoint;
		if(!access(spOutputPoint,m_spOutputAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spOutputNormal;
		if(!access(spOutputNormal,m_spOutputAccessible,
				e_point,e_normal)) return;

		sp<SurfaceAccessorI> spOutputUV;
		if(!access(spOutputUV,m_spOutputAccessible,
				e_point,"uv")) return;

		sp<SurfaceAccessorI> spOutputPartition;
		if(!access(spOutputPartition,m_spOutputAccessible,
				e_primitive,partitionAttr)) return;

		sp<SurfaceAccessorI> spInputVertices;
		if(!access(spInputVertices,m_spInputAccessible,
				e_primitive,e_vertices)) return;

		sp<SurfaceAccessorI> spInputPoint;
		if(!access(spInputPoint,m_spInputAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spInputNormal;
		if(!access(spInputNormal,m_spInputAccessible,
				e_point,e_normal)) return;

		sp<SurfaceAccessorI> spInputUV;
		if(!access(spInputUV,m_spInputAccessible,
				e_point,"uv")) return;

		sp<SurfaceAccessorI> spInputPartition;
		if(iteration && !access(spInputPartition,m_spInputAccessible,
				e_primitive,partitionAttr)) return;

		//* Catmull-Clark

		const U32 primitiveCountIn=spInputVertices->count();
		const U32 pointCountIn=spInputPoint->count();

#if FE_SDO_DEBUG
		feLog("SubdivideOp::handle input points %d primitives %d\n",
				pointCountIn,primitiveCountIn);
#endif

		m_spProfileResize->replace(m_spProfileAccess);

		clearData();

		m_facePoint.resize(primitiveCountIn);

		m_faceSum.resize(pointCountIn,SpatialVector(0.0,0.0,0.0));

		m_faceCount.resize(pointCountIn,0);

		m_midpointSum.resize(pointCountIn,SpatialVector(0.0,0.0,0.0));

		m_midpointCount.resize(pointCountIn,0);

		m_boundarySum.resize(pointCountIn,SpatialVector(0.0,0.0,0.0));

		m_boundaryCount.resize(pointCountIn,0);

		m_spProfileSlot->replace(m_spProfileResize);

		SpatialVector pointsIn[pointCountIn];
		SpatialVector normalsIn[pointCountIn];
		SpatialVector uvsIn[pointCountIn];

		m_pointsIn=pointsIn;
		m_normalsIn=normalsIn;
		m_uvsIn=uvsIn;

		//* make slots for original points (with adjustments)
		for(U32 pointIndexIn=0;pointIndexIn<pointCountIn;pointIndexIn++)
		{
			pointsIn[pointIndexIn]=spInputPoint->spatialVector(pointIndexIn);
			normalsIn[pointIndexIn]=spInputNormal->spatialVector(pointIndexIn);
			uvsIn[pointIndexIn]=spInputUV->spatialVector(pointIndexIn);

			spOutputPoint->append();
			spOutputNormal->set(pointIndexIn,
					spInputNormal->spatialVector(pointIndexIn));
			spOutputUV->set(pointIndexIn,
					spInputUV->spatialVector(pointIndexIn));
		}

		U32 primVertInc=4096;
		U32 primVertSize=primitiveCountIn*5;	//* probably overshoot
		std::vector< std::vector<I32> > primVerts(primVertSize);
		U32 primVertCount=0;

		m_spProfileFace->replace(m_spProfileSlot);

		for(U32 primitiveIndexIn=0;primitiveIndexIn<primitiveCountIn;
				primitiveIndexIn++)
		{
			String partition;
			if(iteration)
			{
				partition=spInputPartition->string(primitiveIndexIn);
			}
			else
			{
				partition.sPrintf("%d",primitiveIndexIn);
			}

			SpatialVector pointSum(0.0,0.0,0.0);
			SpatialVector normalSum(0.0,0.0,0.0);
			SpatialVector uvSum(0.0,0.0,0.0);

			//* UV wrap-around
			Vector2i highUV(false,false);
			SpatialVector lowUV(0.0,0.0,0.0);

			const U32 subCount=spInputVertices->subCount(primitiveIndexIn);

			I32 pointIndices[subCount];

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndexIn=
						spInputVertices->integer(primitiveIndexIn,subIndex);
				pointIndices[subIndex]=pointIndexIn;

				const SpatialVector uv=uvsIn[pointIndexIn];
				pointSum+=pointsIn[pointIndexIn];
				normalSum+=normalsIn[pointIndexIn];
				uvSum+=uv;

				if(uv[0]>0.7)
				{
					highUV[0]=true;
				}
				if(uv[1]>0.7)
				{
					highUV[1]=true;
				}
				lowUV[0]+=(uv[0]<0.3);
				lowUV[1]+=(uv[1]<0.3);
			}

			if(highUV[0])
			{
				uvSum[0]+=lowUV[0];
			}
			if(highUV[1])
			{
				uvSum[1]+=lowUV[1];
			}

			if(subCount)
			{
				const Real scalar=1.0/Real(subCount);

				pointSum*=scalar;
				uvSum*=scalar;

				normalSum=unitSafe(normalSum);
			}

			if(uvSum[0]>=1.0)
			{
				uvSum[0]-=1.0;
			}
			if(uvSum[0]>=1.0)
			{
				uvSum[0]-=1.0;
			}

			m_spProfileFacePoint->replace(m_spProfileFace);

			//* face point
			const I32 pointIndexOut=spOutputPoint->append();

			spOutputPoint->set(pointIndexOut,pointSum);
			spOutputNormal->set(pointIndexOut,normalSum);
			spOutputUV->set(pointIndexOut,uvSum);

			m_facePoint[primitiveIndexIn]=pointIndexOut;

			if(primVertCount+subCount>=primVertSize)
			{
				m_spProfileVertSize->replace(m_spProfileFacePoint);

				primVertSize+=primVertInc;
				primVerts.resize(primVertSize);

				m_spProfileFacePoint->replace(m_spProfileVertSize);
			}

			m_spProfileEdgeMap->replace(m_spProfileFacePoint);

			I32 wrapIndex= -1;
			I32 lastEdgePointIndex= -1;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				//* edge face
				const I32 pointIndexA=pointIndices[subIndex];
				const I32 pointIndexB=pointIndices[(subIndex+1)%subCount];

				Edge& rEdge=(pointIndexA>pointIndexB)?
						m_edgeMap[(U64(pointIndexA)<<32)+pointIndexB]:
						m_edgeMap[(U64(pointIndexB)<<32)+pointIndexA];

				rEdge.m_faceSet.insert(primitiveIndexIn);

				m_spProfileEdgePoint->replace(m_spProfileEdgeMap);

				if(rEdge.m_pointIndex<0)
				{
					rEdge.m_pointIndex=spOutputPoint->append();
				}

				m_spProfileVertex->replace(m_spProfileEdgePoint);

				//* face influence on pre-existing points
				m_faceSum[pointIndexA]+=pointSum;
				m_faceCount[pointIndexA]++;

				//* create sub primitive
#if FALSE
				const I32 primitiveIndexOut=spOutputVertices->append();

				spOutputPartition->set(primitiveIndexOut,partition);

				//* vertex 0: original point
				spOutputVertices->append(primitiveIndexOut,pointIndexA);

				//* vertex 1: midpoint
				spOutputVertices->append(primitiveIndexOut,rEdge.m_pointIndex);

				//* vertex 2: face point
				spOutputVertices->append(primitiveIndexOut,pointIndexOut);

				//* vertex 3: prev midpoint
				spOutputVertices->append(primitiveIndexOut,lastEdgePointIndex);

				if(subIndex==subCount-1)
				{
					spOutputVertices->set(wrapIndex,3,rEdge.m_pointIndex);
				}
				if(!subIndex)
				{
					wrapIndex=primitiveIndexOut;
				}
#else
				std::vector<I32>& rVerts=primVerts[primVertCount++];
				rVerts.resize(4);

				const I32 primitiveIndexOut=spOutputVertices->count();

				//* TODO
//				spOutputPartition->set(primitiveIndexOut,partition);

				//* vertex 0: original point
				rVerts[0]=pointIndexA;

				//* vertex 1: midpoint
				rVerts[1]=rEdge.m_pointIndex;

				//* vertex 2: face point
				rVerts[2]=pointIndexOut;

				//* vertex 3: prev midpoint
				rVerts[3]=lastEdgePointIndex;

				if(subIndex==subCount-1)
				{
					primVerts[wrapIndex][3]=rEdge.m_pointIndex;
				}
				if(!subIndex)
				{
					wrapIndex=primVertCount-1;
				}
#endif

				lastEdgePointIndex=rEdge.m_pointIndex;

				m_spProfileEdgeMap->replace(m_spProfileVertex);
			}

			m_spProfileFace->replace(m_spProfileEdgeMap);
		}

		m_spProfileAppend->replace(m_spProfileFace);

		primVerts.resize(primVertCount);
		spOutputVertices->append(primVerts);

		m_spProfileEdge->replace(m_spProfileAppend);

		//* set edge points
		adjustThreads();	//* in case auto was just turned on
		const U32 threadCount=catalog<I32>("Threads");
		setNonAtomicRange(0,threadCount-1);
		setStage("edge");
		OperatorThreaded::handle(a_rSignal);

		m_spProfilePoint->replace(m_spProfileEdge);

		//* adjust original points
		setAtomicRange(m_spInputAccessible,e_pointsOnly);
		setStage("point");
		OperatorThreaded::handle(a_rSignal);

		m_spProfileAccess->replace(m_spProfilePoint);

#if FE_SDO_DEBUG
		feLog("SubdivideOp::handle"
				" iteration %d subdivided points %d primitives %d\n",
				iteration,spOutputPoint->count(),spOutputVertices->count());
#endif
	}

	if(recomputeNormals)
	{
		m_spProfileNormal->replace(m_spProfileAccess);

		sp<Catalog> spOtherCatalog=m_spSurfaceNormalOp;
		if(spOtherCatalog.isValid())
		{
			spOtherCatalog->catalog<I32>("Threads")=
					catalog<I32>("Threads");
			spOtherCatalog->catalog<bool>("AutoThread")=
					catalog<bool>("AutoThread");
			spOtherCatalog->catalog<bool>("Paging")=
					catalog<bool>("Paging");
			spOtherCatalog->catalog<bool>("StayAlive")=
					catalog<bool>("StayAlive");
		}

		m_spSurfaceNormalOp->handle(a_rSignal);

		m_spProfileAccess->replace(m_spProfileNormal);
	}

	clearData();

	m_spProfileAccess->finish();
	m_spProfiler->end();

#if FE_SDO_PROFILE
	feLog("%s:\n%s\n",m_spProfiler->name().c_str(),
			m_spProfiler->report().c_str());
#endif
}

void SubdivideOp::run(I32 a_id,sp<SpannedRange> a_spRange,String a_stage)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::run id %d stage \"%s\" range %s\n",
			a_id,a_stage.c_str(),a_spRange->brief().c_str());
#endif

	if(a_stage=="edge")
	{
		run_edge(a_id,a_spRange);
	}
	else if(a_stage=="point")
	{
		run_point(a_id,a_spRange);
	}
}

void SubdivideOp::run_edge(I32 a_id,sp<SpannedRange> a_spRange)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::run_edge id %d\n",a_id);
#endif

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputNormal;
	if(!access(spOutputNormal,m_spOutputAccessible,e_point,e_normal)) return;

	sp<SurfaceAccessorI> spOutputUV;
	if(!access(spOutputUV,m_spOutputAccessible,e_point,"uv")) return;

	sp<SurfaceAccessorI> spInputPoint;
	if(!access(spInputPoint,m_spInputAccessible,e_point,e_position)) return;

	//* assuming contiguous
	SpannedRange::Iterator it=a_spRange->begin();
	if(it.atEnd())
	{
		return;
	}
	const U32 modulus=it.value();
	const U32 threadCount=catalog<I32>("Threads");
	const U32 pointCountIn=spInputPoint->count();

#if TRUE
	const U32 pointMin=modulus*pointCountIn/threadCount;
	const U32 pointMax=(modulus==threadCount-1)? pointCountIn-1:
			(modulus+1)*pointCountIn/threadCount-1;
#else
	if(modulus)
	{
		return;
	}
	const U32 pointMin=0;
	const U32 pointMax=pointCountIn-1;
#endif

//	feLog("mod %d/%d count %d span %d %d\n",
//			modulus,threadCount,pointCountIn,pointMin,pointMax);

	for(std::map<U64,Edge>::iterator it=m_edgeMap.begin();
			it!=m_edgeMap.end();it++)
	{
		const U64 combo=it->first;
		const Edge& rEdge=it->second;

		const I32 pointIndexA=combo>>32;
		const I32 pointIndexB=combo&0xFFFFFFFF;

		const BWORD changeA=(pointIndexA>=pointMin && pointIndexA<=pointMax);
		const BWORD changeB=(pointIndexB>=pointMin && pointIndexB<=pointMax);

		if(!changeA && !changeB)
		{
			continue;
		}

		const std::set<I32>& rFaceSet=rEdge.m_faceSet;

		const SpatialVector pointA=m_pointsIn[pointIndexA];
		const SpatialVector normalA=m_normalsIn[pointIndexA];
		const SpatialVector uvA=m_uvsIn[pointIndexA];

		const SpatialVector pointB=m_pointsIn[pointIndexB];
		const SpatialVector normalB=m_normalsIn[pointIndexB];
		const SpatialVector uvB=m_uvsIn[pointIndexB];

		SpatialVector pointSum=pointA+pointB;
		SpatialVector normalSum=normalA+normalB;
		SpatialVector uvSum=uvA+uvB;

		const SpatialVector midpoint=0.5*pointSum;
		I32 divisor=2;

		//* UV wrap-around
		Vector2i highUV(false,false);
		SpatialVector lowUV(0.0,0.0,0.0);
		if(uvA[0]>0.7 || uvB[0]>0.7)
		{
			highUV[0]=true;
		}
		if(uvA[1]>0.7 || uvB[1]>0.7)
		{
			highUV[1]=true;
		}
		lowUV[0]+=(uvA[0]<0.3)+(uvB[0]<0.3);
		lowUV[1]+=(uvA[1]<0.3)+(uvB[1]<0.3);

		if(!m_flat && rFaceSet.size()>1)
		{
			//* also add adjacent face points
			for(std::set<I32>::iterator it=rFaceSet.begin();
					it!=rFaceSet.end();it++)
			{
				const I32 primitiveIndexIn=*it;
				const I32 pointIndexOut=m_facePoint[primitiveIndexIn];
				const SpatialVector uv=
						spOutputUV->spatialVector(pointIndexOut);

				pointSum+=spOutputPoint->spatialVector(pointIndexOut);
				normalSum+=spOutputNormal->spatialVector(pointIndexOut);
				uvSum+=uv;
				divisor++;

				if(uv[0]>0.7)
				{
					highUV[0]=true;
				}
				if(uv[1]>0.7)
				{
					highUV[1]=true;
				}
				lowUV[0]+=(uv[0]<0.3);
				lowUV[1]+=(uv[1]<0.3);
			}

			if(changeA)
			{
				m_midpointSum[pointIndexA]+=midpoint;
				m_midpointCount[pointIndexA]++;
			}

			if(changeB)
			{
				m_midpointSum[pointIndexB]+=midpoint;
				m_midpointCount[pointIndexB]++;
			}
		}
		else
		{
			if(changeA)
			{
				m_boundarySum[pointIndexA]+=midpoint;
				m_boundaryCount[pointIndexA]++;
			}

			if(changeB)
			{
				m_boundarySum[pointIndexB]+=midpoint;
				m_boundaryCount[pointIndexB]++;
			}
		}

		if(highUV[0])
		{
			uvSum[0]+=lowUV[0];
		}
		if(highUV[1])
		{
			uvSum[1]+=lowUV[1];
		}

		const Real scalar=1.0/Real(divisor);

		pointSum*=scalar;
		uvSum*=scalar;

		normalSum=unitSafe(normalSum);

		if(uvSum[0]>=1.0)
		{
			uvSum[0]-=1.0;
		}
		if(uvSum[0]>=1.0)
		{
			uvSum[0]-=1.0;
		}

		//* NOTE  rEdge.m_pointIndex isn't pointIndexA,
		//* but this should keep it exclusive
		if(changeA)
		{
			spOutputPoint->set(rEdge.m_pointIndex,pointSum);
			spOutputNormal->set(rEdge.m_pointIndex,normalSum);
			spOutputUV->set(rEdge.m_pointIndex,uvSum);
		}
	}
}

void SubdivideOp::run_point(I32 a_id,sp<SpannedRange> a_spRange)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::run_point id %d\n",a_id);
#endif

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,
			e_point,e_position)) return;

	if(m_flat)
	{
		for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
		{
			const U32 pointIndexIn=it.value();

			spOutputPoint->set(pointIndexIn,m_pointsIn[pointIndexIn]);
		}
	}
	else
	{
		for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
		{
			const U32 pointIndexIn=it.value();

			const SpatialVector pointIn=m_pointsIn[pointIndexIn];

			const I32 pointFaceCount=m_faceCount[pointIndexIn];
			FEASSERT(pointFaceCount>0);

			const SpatialVector faceAverage=
					m_faceSum[pointIndexIn]/Real(pointFaceCount);

			if(pointFaceCount>2)
			{
				const I32 divisor=m_midpointCount[pointIndexIn]+
						m_boundaryCount[pointIndexIn];
				FEASSERT(divisor>0);
				const SpatialVector midpointAverage=
						(m_midpointSum[pointIndexIn]+
						m_boundarySum[pointIndexIn])/
						Real(divisor);

				const SpatialVector pointOut=
						(faceAverage+2.0*midpointAverage+
						(pointFaceCount-3)*pointIn)/Real(pointFaceCount);

				spOutputPoint->set(pointIndexIn,pointOut);
			}
			else
			{
				//* TODO find out what everyone else does on boundaries
				FEASSERT(m_boundaryCount[pointIndexIn]>0);
				const SpatialVector boundaryAverage=
						m_boundarySum[pointIndexIn]/
						Real(m_boundaryCount[pointIndexIn]);

				const SpatialVector pointOut=
						0.5*(boundaryAverage+pointIn);

				spOutputPoint->set(pointIndexIn,pointOut);
			}
		}
	}
}

} // namespace
