/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_SS_DEBUG				FALSE
#define	FE_SS_TREE_HIT_DEBUG	FALSE

namespace fe
{

SurfaceSearchable::SurfaceSearchable(void):
	m_refinement(0),
	m_elements(0),
	m_vertices(0),
	m_pElementArray(NULL),
	m_pVertexArray(NULL),
	m_pNormalArray(NULL),
	m_pUVArray(NULL),
	m_pUV3Array(NULL),
	m_pTriangleIndexArray(NULL),
	m_pPartitionIndexArray(NULL),
	m_pPointIndexArray(NULL),
	m_pPrimitiveIndexArray(NULL)
{
	m_radius=0.0;
}

SurfaceSearchable::~SurfaceSearchable(void)
{
	deallocate(m_pPartitionIndexArray);
	deallocate(m_pPrimitiveIndexArray);
	deallocate(m_pPointIndexArray);
	deallocate(m_pTriangleIndexArray);
	deallocate(m_pUV3Array);
	deallocate(m_pUVArray);
	deallocate(m_pNormalArray);
	deallocate(m_pVertexArray);
	deallocate(m_pElementArray);
}

void SurfaceSearchable::resizeArrays(U32 a_elementCount,U32 a_vertexCount)
{
	m_pElementArray=(Vector3i*)reallocate(m_pElementArray,
			a_elementCount*sizeof(Vector3i));

	m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
			a_vertexCount*sizeof(SpatialVector));
	m_pNormalArray=(SpatialVector*)reallocate(m_pNormalArray,
			a_vertexCount*sizeof(SpatialVector));
	m_pUVArray=(Vector2*)reallocate(m_pUVArray,
			a_vertexCount*sizeof(Vector2));
	m_pTriangleIndexArray=(I32*)reallocate(m_pTriangleIndexArray,
			a_vertexCount*sizeof(I32));
	m_pPointIndexArray=(I32*)reallocate(m_pPointIndexArray,
			a_vertexCount*sizeof(I32));
	m_pPrimitiveIndexArray=(I32*)reallocate(m_pPrimitiveIndexArray,
			a_vertexCount*sizeof(I32));
	m_pPartitionIndexArray=(I32*)reallocate(m_pPartitionIndexArray,
			a_vertexCount*sizeof(I32));
}

void SurfaceSearchable::resizeFor(U32 a_primitiveCount,U32 a_vertexCount,
	U32& a_rElementAllocated,U32& a_rVertexAllocated)
{
	U32 elementDesired=a_rElementAllocated;
#if FALSE
	if(!a_rElementAllocated)
	{
		elementDesired=a_vertexCount*a_primitiveCount/3;
	}
	else if(a_rElementAllocated<a_vertexCount)
	{
		elementDesired=(a_vertexCount*1.2)/3;	//* tweak
	}
#else
	if(elementDesired<a_primitiveCount)
	{
		elementDesired=a_primitiveCount;
	}
#endif

	U32 vertexDesired=a_rVertexAllocated;
	if(!a_rVertexAllocated)
	{
		vertexDesired=a_vertexCount*a_primitiveCount;
	}
	else if(a_rVertexAllocated<a_vertexCount)
	{
		vertexDesired=a_vertexCount*1.2;	//* tweak
	}

	//* don't shrink here
	if(elementDesired<a_rElementAllocated)
	{
		elementDesired=a_rElementAllocated;
	}
	if(vertexDesired<a_rVertexAllocated)
	{
		vertexDesired=a_rVertexAllocated;
	}

	if(a_rElementAllocated!=elementDesired ||
			a_rVertexAllocated!=vertexDesired)
	{
		resizeArrays(elementDesired,vertexDesired);
		a_rElementAllocated=elementDesired;
		a_rVertexAllocated=vertexDesired;
	}
}


Protectable* SurfaceSearchable::clone(Protectable* pInstance)
{
#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::clone\n");
#endif

	SurfaceSearchable* pSurfaceSearchable= pInstance?
			dynamic_cast<SurfaceSearchable*>(pInstance):
			new SurfaceSearchable();

	checkCache();

	SurfaceSphere::clone(pSurfaceSearchable);

	// TODO copy attributes

	pSurfaceSearchable->m_elements=m_elements;
	pSurfaceSearchable->m_vertices=m_vertices;
	pSurfaceSearchable->resizeArrays(m_elements,m_vertices);

	const U32 sizeVector=m_vertices*sizeof(SpatialVector);
	const U32 sizeVector2=m_vertices*sizeof(Vector2);
	const U32 sizeVector3i=m_vertices*sizeof(Vector3i);
	const U32 sizeI32=m_vertices*sizeof(I32);

	memcpy(pSurfaceSearchable->m_pElementArray,
			m_pElementArray,sizeVector3i);
	memcpy(pSurfaceSearchable->m_pVertexArray,
			m_pVertexArray,sizeVector);
	memcpy(pSurfaceSearchable->m_pNormalArray,
			m_pNormalArray,sizeVector);
	memcpy(pSurfaceSearchable->m_pTriangleIndexArray,
			m_pTriangleIndexArray,sizeI32);
	memcpy(pSurfaceSearchable->m_pPointIndexArray,
			m_pPointIndexArray,sizeI32);
	memcpy(pSurfaceSearchable->m_pPrimitiveIndexArray,
			m_pPrimitiveIndexArray,sizeI32);
	memcpy(pSurfaceSearchable->m_pPartitionIndexArray,
			m_pPartitionIndexArray,sizeI32);

	if(pSurfaceSearchable->m_pUVArray)
	{
		memcpy(pSurfaceSearchable->m_pUVArray,m_pUVArray,sizeVector2);
	}
	else
	{
		deallocate(m_pUVArray);
		m_pUVArray=NULL;
	}

	return pSurfaceSearchable;
}

void SurfaceSearchable::cache(void)
{
	feLog("SurfaceSearchable::cache\n");

	if(!m_record.isValid())
	{
		feLog("SurfaceSearchable::cache record invalid\n");
		return;
	}

	SurfaceSphere::cache();

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_record);

	sp<RecordGroup> spRG=surfaceModelRV.surfacePointRG();
	for(RecordGroup::iterator it=spRG->begin();it!=spRG->end();it++)
	{
		sp<RecordArray> spRA= *it;

		SurfacePoint m_surfacePointRV;
		m_surfacePointRV.bind(spRA);
		while(m_surfacePointRV.step())
		{
			SpatialVector& at=m_surfacePointRV.at();
			SpatialVector& up=m_surfacePointRV.up();
//			SpatialVector& uvw=m_surfacePointRV.uvw();

			resizeArrays(m_elements,m_vertices);

//			Vector3i& rPrimitive=m_pElementArray[m_elements];
			SpatialVector& rVertex=m_pVertexArray[m_vertices];
			SpatialVector& rNormal=m_pNormalArray[m_vertices];
			Vector2& rUV=m_pUVArray[m_vertices];
			I32& rTriangleIndex=m_pTriangleIndexArray[m_vertices];
			I32& rPointIndex=m_pPointIndexArray[m_vertices];
			I32& rPrimitiveIndex=m_pPrimitiveIndexArray[m_vertices];
			I32& rPartitionIndex=m_pPartitionIndexArray[m_vertices];

			rVertex=at;
			rNormal=up;
			//* TODO element array
			//* TODO uv, triangle, primitive, partition

			printf("vertex %d v %s n %s uv %s"
					" triangle %d point %d primitive %d partition %d\n",
					m_vertices,c_print(rVertex),c_print(rNormal),
					c_print(rUV),rTriangleIndex,
					rPointIndex, rPrimitiveIndex,rPartitionIndex);
			m_vertices++;
		}
	}

	if(surfaceModelRV.searchComponent().isValid())
	{
		m_spSearchComponent=surfaceModelRV.searchComponent();
	}
}

void SurfaceSearchable::filterUV(void)
{
	if(!m_pUVArray)
	{
		return;
	}

	for(U32 elementIndex=0;elementIndex<m_elements;elementIndex++)
	{
		const U32 subCount=m_pElementArray[elementIndex][1];
		const BWORD openCurve=m_pElementArray[elementIndex][2];
		if(subCount!=3 || openCurve)
		{
			//* TODO
			continue;
		}

		//* filter UV wrap-around on each triangle

		const U32 pointIndex=m_pElementArray[elementIndex][0];

		FEASSERT(m_pUVArray);
		Vector2& rUV0=m_pUVArray[pointIndex];
		Vector2& rUV1=m_pUVArray[pointIndex+1];
		Vector2& rUV2=m_pUVArray[pointIndex+2];

//			feLog("filter %d/%d  %s  %s  %s\n",pointIndex,m_vertices,
//					c_print(rUV0),c_print(rUV1),c_print(rUV2));

		for(U32 m=0;m<2;m++)
		{
			if(rUV0[m]<0.3 && (rUV1[m]>0.7 || rUV2[m]>0.7))
			{
				rUV0[m]+=1.0;
			}

			if(rUV1[m]<0.3 && (rUV0[m]>0.7 || rUV2[m]>0.7))
			{
				rUV1[m]+=1.0;
			}

			if(rUV2[m]<0.3 && (rUV0[m]>0.7 || rUV1[m]>0.7))
			{
				rUV2[m]+=1.0;
			}
		}

//		feLog("filtered %s  %s  %s\n",
//				c_print(rUV0),c_print(rUV1),c_print(rUV2));
	}
}

void SurfaceSearchable::prepareForSample(void)
{
	SurfaceBase::prepareForSample();

	filterUV();
}

void SurfaceSearchable::prepareForSearch(void)
{
	checkCache();

	if(m_spSpatialTree.isValid())
	{
		return;
	}

#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::prepareForSearch"
			" m_elements=%d m_vertices=%d tree valid=%d\n",
			m_elements,m_vertices,m_spSpatialTree.isValid());
#endif

	if(!m_elements)
	{
		feX(e_cannotCreate,"SurfaceSearchable::prepareForSearch","no elements");
	}
	if(!m_vertices)
	{
		feX(e_cannotCreate,"SurfaceSearchable::prepareForSearch","no vertices");
	}
	if(m_spSearchComponent.isValid())
	{
		m_spSpatialTree=m_spSearchComponent;
	}
	if(!m_spSpatialTree.isValid())
	{
		m_spSpatialTree=create("SpatialTreeI.default");
	}
	if(!m_spSpatialTree.isValid())
	{
		return;
	}

#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::prepareForSearch building spatial tree\n");
#endif

	filterUV();

	m_spSpatialTree->setRefinement(m_refinement);
	m_spSpatialTree->populate(m_pElementArray,m_pVertexArray,m_pNormalArray,
			m_pUVArray,m_pTriangleIndexArray, m_pPointIndexArray,
			m_pPrimitiveIndexArray,m_pPartitionIndexArray,
			m_elements,m_vertices,center(),radius());
}

void SurfaceSearchable::prepareForUVSearch(void)
{
	checkCache();

	if(!m_pUVArray)
	{
		feX(e_cannotFind,"SurfaceSearchable::prepareForUVSearch","no UVs");
	}

	prepareForSearch();

	if(m_spUVTree.isValid())
	{
		return;
	}

#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::prepareForUVSearch"
			" m_elements=%d m_vertices=%d tree valid=%d\n",
			m_elements,m_vertices,m_spUVTree.isValid());
#endif

	if(!m_elements || !m_vertices)
	{
		return;
	}
	if(!m_spUVTree.isValid())
	{
		m_spUVTree=create("SpatialTreeI.QuadTree");
	}
	if(!m_spUVTree.isValid())
	{
		return;
	}

#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::prepareForSearchByUV building spatial tree\n");
#endif

	m_pUV3Array=(SpatialVector*)reallocate(m_pUV3Array,
			m_vertices*sizeof(SpatialVector));
	for(U32 index=0;index<m_vertices;index++)
	{
		m_pUV3Array[index]=m_pUVArray[index];

#if	FE_SS_DEBUG
		feLog("  prep %d triangle %d point %d primitive %d partition %d"
				" uv %s  %s\n",
				index,
				m_pTriangleIndexArray[index],
				m_pPointIndexArray[index],
				m_pPrimitiveIndexArray[index],
				m_pPartitionIndexArray[index],
				c_print(m_pUVArray[index]),
				c_print(m_pUV3Array[index]))
#endif
	}

	m_spUVTree->setRefinement(0);
	m_spUVTree->populate(m_pElementArray,m_pUV3Array,m_pNormalArray,
			m_pUVArray,m_pTriangleIndexArray, m_pPointIndexArray,
			m_pPrimitiveIndexArray,m_pPartitionIndexArray,
			m_elements,m_vertices,SpatialVector(0.0,0.0,0.0),0.0);
}

SpatialTransform SurfaceSearchable::sample(Vector2 a_uv) const
{
//	feLog("SurfaceSearchable::sample uv %s\n",c_print(a_uv));

	sp<SurfaceSearchable::ImpactI> spImpact=nearestPoint(a_uv);
	if(spImpact.isNull())
	{
		feLog("SurfaceSearchable::sample spImpact invalid\n");
		SpatialTransform result;
		setIdentity(result);
		return result;
	}

	const I32 tri=spImpact->triangleIndex();
	const SpatialBary barycenter=spImpact->barycenter();

//	feLog("SurfaceSearchable::sample uv %s tri %d bary %s\n",
//			c_print(a_uv),tri,c_print(barycenter));

	return reinterpret_cast<const SurfaceBase* const>(this)->sample(
			tri,barycenter);
}

sp<SurfaceSearchable::ImpactI> SurfaceSearchable::nearestPoint(
		const Vector2& a_uv) const
{
	if(!m_elements || !m_vertices || !m_spUVTree.isValid())
	{
		feLog("SurfaceSearchable::nearestPoint(%s) verts %d tree %d\n",
				c_print(a_uv),m_vertices,m_spUVTree.isValid());

		return sp<ImpactI>(NULL);
	}

	SpatialVector origin(a_uv);
	const Real maxDistance= -1.0;
	const U32 hitLimit=1;
	const BWORD anyHit=FALSE;
	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	m_spUVTree->nearestPoint(origin,maxDistance,anyHit,hitLimit,
			m_spPartition,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
		return sp<ImpactI>(NULL);
	}

	sp<SpatialTreeI::Hit>& rspHit=hitArray[0];

	sp<Impact> spImpact=m_pImpactPool->get();
	spImpact->copy(rspHit);
	spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
	spImpact->setOrigin(origin);

	//* TODO reset vertices to original locations, not UV substitutes

	return spImpact;
}

void SurfaceSearchable::Impact::copy(sp<SpatialTreeI::Hit>& a_rspHit)
{
	setUV(a_rspHit->uv());
	setDu(a_rspHit->du());
	setDv(a_rspHit->dv());
	setBarycenter(a_rspHit->barycenter());
	setTriangleIndex(a_rspHit->triangleIndex());
	setPrimitiveIndex(a_rspHit->primitiveIndex());
	setPartitionIndex(a_rspHit->partitionIndex());
	setDirection(a_rspHit->direction());
	setDistance(a_rspHit->distance());
	setIntersectionLocal(a_rspHit->intersection());

	//* should be done in resolveImpact
//	const SpatialVector* normalArray=a_rspHit->normal();
//	const SpatialVector weightedNormal=fe::location(a_rspHit->barycenter(),
//			normalArray[0],normalArray[1],normalArray[2]);
//	setNormalLocal(unitSafe(weightedNormal));
}

std::vector< sp<SurfaceI::ImpactI> > SurfaceSearchable::nearestPoints(
		const SpatialVector& a_origin,Real a_maxDistance,U32 a_hitLimit) const
{
	std::vector< sp<SurfaceI::ImpactI> > impactArray;

	if(!m_elements || !m_vertices || !m_spSpatialTree.isValid())
	{
		return impactArray;
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const BWORD anyHit=FALSE;
	m_spSpatialTree->nearestPoint(a_origin,a_maxDistance,anyHit,a_hitLimit,
			m_spPartition,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
#if FE_SS_DEBUG
		feLog("SurfaceSearchable::nearestPoints no hits\n");
#endif
		return impactArray;
	}

	impactArray.resize(hitCount);

	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		sp<Impact> spImpact=m_pImpactPool->get();
		spImpact->copy(rspHit);
		spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
		spImpact->setOrigin(a_origin);
		impactArray[m]=spImpact;
	}

	return impactArray;
}

sp<SurfaceSearchable::ImpactI> SurfaceSearchable::nearestPoint(
		const SpatialVector& a_origin,Real a_maxDistance,BWORD a_anyHit) const
{
	if(!m_elements || !m_vertices || !m_spSpatialTree.isValid())
	{
		return sp<ImpactI>(NULL);
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const U32 hitLimit=1;	//* TODO param multi-weighting
	m_spSpatialTree->nearestPoint(a_origin,a_maxDistance,a_anyHit,hitLimit,
			m_spPartition,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
#if FE_SS_DEBUG
		feLog("SurfaceSearchable::nearestPoint no hits\n");
#endif
		return sp<ImpactI>(NULL);
	}

	sp<Impact> spImpact0=m_pImpactPool->get();
	spImpact0->copy(hitArray[0]);
	spImpact0->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
	spImpact0->setOrigin(a_origin);

	return spImpact0;
	//* TODO clean up blending is still needed

#if FE_SS_TREE_HIT_DEBUG
	feLog("\n");
	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		feLog("Hit %d/%d dist %.6G  inter %s\n", m,hitCount,
				rspHit->distance(),c_print(rspHit->intersection()));
		feLog("  face %d triangleIndex %d bary %s\n",rspHit->face(),
				rspHit->triangleIndex(),c_print(rspHit->barycenter()));
		for(U32 n=0;n<3;n++)
		{
			feLog("  v%d %s n %s\n",n,c_print(rspHit->vertex()[n]),
					c_print(rspHit->normal()[n]));
/*
			FEASSERT(fabs(magnitude(rspHit->vertex()[n]))<1e9);
			FEASSERT(fabs(magnitude(rspHit->normal()[n]))<1.1);
			if(fabs(magnitude(rspHit->vertex()[n]))>1e9 ||
					fabs(magnitude(rspHit->normal()[n]))>1.1)
			{
				exit(1);
			}
*/
		}
	}
#endif

	Real hitWeight[hitLimit];
	hitWeight[0]=1.0;

#if TRUE
	//* simply closest
	const Real sum=1.0;
	for(U32 m=1;m<hitCount;m++)
	{
		hitWeight[m]=0.0;
	}
#endif

	// TODO pick single best hit
	U32 bestHit=0;
	Real bestScore=0.0;
	for(U32 m=0;m<hitCount;m++)
	{
		const sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		Real thisScore=0.0;
		const SpatialVector diff=a_origin-rspHit->intersection();
		const Real mag=magnitude(diff);
		if(mag<1e-6)
		{
			thisScore=1.0;
		}
		else
		{
			const SpatialVector to=diff/mag;
			const SpatialVector norm=unitSafe(location(rspHit->barycenter(),
					rspHit->normal()[0],rspHit->normal()[1],
					rspHit->normal()[2]));
			const Real to_dot=dot(to,norm);
			thisScore=(to_dot>0.0)? to_dot: 0.0;
#if FE_SS_TREE_HIT_DEBUG
			feLog("  to %s\n",c_print(to));
#endif
		}
		if(!m)
		{
			bestScore=thisScore;
		}
		else if(thisScore>bestScore)
		{
			bestScore=thisScore;
			bestHit=m;
		}
#if FE_SS_TREE_HIT_DEBUG
		feLog("Score %d/%d %.6G best %.6G\n",
				m,hitCount,thisScore,bestScore);
#endif
	}
	//* single weighted
	for(U32 m=0;m<hitCount;m++)
	{
		hitWeight[m]=0.0;
	}
	hitWeight[bestHit]=1.0;

	sp<SpatialTreeI::Hit>& rspHit=hitArray[bestHit];
	sp<Impact> spImpact=m_pImpactPool->get();
	spImpact->copy(rspHit);
	spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
	spImpact->setOrigin(a_origin);

#if FALSE
	//* reevaluate barycenter based on normals
//	const SpatialVector to=unit(a_origin-rspHit->intersection());
	const SpatialVector norm=unitSafe(location(rspHit->barycenter(),
			rspHit->normal()[0],rspHit->normal()[1],rspHit->normal()[2]));
	SpatialBary barynormal;
	barynormal.solve(rspHit->normal()[0],rspHit->normal()[1],
			rspHit->normal()[2],norm);
	spImpact->setBarycenter(barynormal);
	const SpatialVector intersection=location(barynormal,
			rspHit->vertex()[0],rspHit->vertex()[1],rspHit->vertex()[2]);
	const SpatialVector toward=intersection-a_origin;
	const Real range=magnitude(toward);
	spImpact->setDistance(range);
	spImpact->setDirection(range>0.0? toward/range: toward);
#endif

	if(hitCount>1)
	{
		SpatialVector intersection;
		SpatialVector normal;
		set(intersection);
		set(normal);
		for(U32 m=0;m<hitCount;m++)
		{
			const Real fraction=hitWeight[m]/sum;
			intersection+=fraction*hitArray[m]->intersection();

			const SpatialVector* norms=hitArray[m]->normal();
			const Vector2& bary=hitArray[m]->barycenter();
			normal+=fraction*(norms[0]*bary[0]+norms[1]*bary[1]+
					norms[2]*(1.0f-bary[0]-bary[1]));
		}
		normalizeSafe(normal);
		const SpatialVector to=intersection-a_origin;
		const Real mag=magnitude(to);
		spImpact->setDistance(mag);
		spImpact->setDirection(mag>1e-6? to/mag: to);

#if FE_SS_TREE_HIT_DEBUG
		feLog("Blend %.6G %s\n",
				spImpact->distance(),c_print(intersection));
#endif
	}

/*
	feLog("dist %.6G dir %s\n",spImpact->distance(),
			c_print(spImpact->direction()));
	feLog("normal0 %s\n",c_print(spImpact->normal0()));
	feLog("normal1 %s\n",c_print(spImpact->normal1()));
	feLog("normal2 %s\n",c_print(spImpact->normal2()));
*/

	return spImpact;
}

std::vector< sp<SurfaceI::ImpactI> > SurfaceSearchable::rayImpacts(
		const SpatialVector& a_origin,const SpatialVector& a_direction,
		Real a_maxDistance,U32 a_hitLimit) const
{
	std::vector< sp<SurfaceI::ImpactI> > impactArray;

	if(!m_elements || !m_vertices || !m_spSpatialTree.isValid())
	{
		return impactArray;
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const Real distance=m_spSpatialTree->rayImpact(a_origin,a_direction,
			a_maxDistance,FALSE,a_hitLimit,m_spPartition,hitArray);
	const U32 hitCount=hitArray.size();

	if(!hitCount || (a_maxDistance>0.0 && distance>a_maxDistance))
	{
		return impactArray;
	}

	impactArray.resize(hitCount);
	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		sp<Impact> spImpact=m_pImpactPool->get();
		spImpact->copy(rspHit);
		spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
		spImpact->setOrigin(a_origin);
		spImpact->setDirection(a_direction);
		impactArray[m]=spImpact;
	}
	return impactArray;
}

sp<SurfaceSearchable::ImpactI> SurfaceSearchable::rayImpact(
		const SpatialVector& a_origin,const SpatialVector& a_direction,
		Real a_maxDistance,BWORD a_anyHit) const
{
	if(!m_elements || !m_vertices || !m_spSpatialTree.isValid())
	{
		return sp<ImpactI>(NULL);
	}

	const U32 hitLimit=1;
	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const Real distance=m_spSpatialTree->rayImpact(a_origin,a_direction,
			a_maxDistance,a_anyHit,hitLimit,m_spPartition,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount || (a_maxDistance>0.0 && distance>a_maxDistance))
	{
		return sp<ImpactI>(NULL);
	}

	sp<SpatialTreeI::Hit>& rspHit=hitArray[0];

	sp<Impact> spImpact=m_pImpactPool->get();
	spImpact->copy(rspHit);
	spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
	spImpact->setOrigin(a_origin);
	spImpact->setDirection(a_direction);
	return spImpact;
}

void SurfaceSearchable::calcBoundingSphere(void)
{
	if(!m_vertices)
	{
		set(m_center);
		m_radius=0.0;
		return;
	}

#if FALSE
	//* by center of "mass"
	m_center=m_pVertexArray[0];
	for(U32 m=1;m<m_vertices;m++)
	{
		m_center+=m_pVertexArray[m];
	}
	m_center*=1.0/m_vertices;
#else
	//* by bounding box
	SpatialVector min=m_pVertexArray[0];
	SpatialVector max=min;
	for(U32 m=1;m<m_vertices;m++)
	{
		SpatialVector& rVertex=m_pVertexArray[m];
		for(U32 n=0;n<3;n++)
		{
			if(min[n]>rVertex[n])
			{
				min[n]=rVertex[n];
			}
			if(max[n]<rVertex[n])
			{
				max[n]=rVertex[n];
			}
		}
	}
	m_center=0.5*(min+max);
#endif

	Real max2=0.0;
	for(U32 m=1;m<m_vertices;m++)
	{
		SpatialVector radial=m_pVertexArray[m]-m_center;
		Real dist2=magnitudeSquared(radial);
		if(max2<dist2)
		{
			max2=dist2;
		}
	}
	m_radius=sqrt(max2);

#if FE_SS_DEBUG
	feLog("SurfaceSearchable::calcBoundingSphere"
			" vertices %d center %s radius %.6G\n",
			m_vertices,c_print(m_center),m_radius);
#endif
}

} // namespace
